////////////////////////////////////////////////////////////////////////////////
//Name:         scmd.c                                                         /
//Description:  Serial command accumulator/processor                           /
//              To be used as serial receive interrupt                         /
//              handler for processors with hardware UARTS.                    /
//              UART overrun error flag must be checked and                    /
//              corrected at main level periodically.                          /
////////////////////////////////////////////////////////////////////////////////
//Taylor Electronics Services (TES)                                            /
//                                                                             /
//  This source code is public domain.                                         /
////////////////////////////////////////////////////////////////////////////////
//Revision history:                                                            /
//2008/05/01    Changed "Byte" to "unsigned int8" because the handling in the  /
//              compiler changed at some point where n>0 started to act on     /
//              variables as signed numbers giving odd responses.              /
//              Added [lf] after [cr] so the terminal prog would be happy      /
//              Added SCMD_INIT() to initialize global variables!              /
//2006/01/30    Added option to make buffer larger                             /
//2004/07/03    Made all references either for port 1 or 2 (Not assumed 1)     /
//              Pulled overflow bit definitions back into scmd                 /
//2004/01/22    Add framing test to command 0                                  /
//2003/11/07    Added 2 character null pad after reset                         /
//              command to allow ack+[cr] to escape.                           /
//              Fixed reference to Overrun bits in main code                   /
//              blocks:  was '876==18Fx52, now separate                        /
//2003/06/20    Added multi stream capability                                  /
//2003/05/29    Cleaned up some structure and added kbhit()                    /
//              test to check for another byte waiting to                      /
//              prevent context switching from making me                       /
//              miss a byte.  Added notes on overrun error                     /
//              ram address to be sure to match target.                        /
//              made command queues non-static to save on                      /
//              ROM use.                                                       /
//2002/12/10    Initial release                                                /
////////////////////////////////////////////////////////////////////////////////
//scmd  [0]     command data length in bytes                                   /
//      [1]     command code                                                   /
//      [2-18]  command data, binary                                           /
////////////////////////////////////////////////////////////////////////////////
//low command processing block                                                 /
//00 report firmware code:                                                     /
//$AA00SS[cr]           $       command frame start                            /
//                      AA      unit address                                   /
//                      00      command code                                   /
//                      SS      checksum (if enabled                           /
//                      [cr]    terminator                                     /
//!AA00ccccvvrrSS[cr]   !       response frame start                           /
//                      AA      unit address                                   /
//                      00      command code                                   /
//                      cccc    core                                           /
//                      vv      variant                                        /
//                      rr      revision                                       /
//                      SS      checksum (if enabled)                          /
//                      [cr]    terminator                                     /
//............................................................................./
//01 change unit address:                                                      /
//$AA01aaSS[cr]         $       command frame start                            /
//                      AA      unit address                                   /
//                      01      command code                                   /
//                      aa      new unit address                               /
//                      SS      checksum (if enabled)                          /
//                      [cr]    terminator                                     /
//!aa01rSS[cr]          !       response frame start                           /
//                      aa      current unit address                           /
//                      01      command code                                   /
//                      r       result a=ack/f=frame error                     /
//                      SS      checksum (if enabled)                          /
//                      [cr]    terminator                                     /
//............................................................................./
//02 toggle checksum use:                                                      /
//$AA02SS[cr]           $       command frame start                            /
//                      AA      unit address                                   /
//                      02      command code                                   /
//                      SS      checksum (if enabled)                          /
//                      [cr]    terminator                                     /
//                                                                             /
//!AA02rSS[cr]          AA      unit address                                   /
//                      02      command code                                   /
//                      r       result a=ack/f=frame error                     /
//                      SS      checksum (if enabled)                          /
//                      [cr]    terminator                                     /
//............................................................................./
//07 software reset:                                                           /
//$AA07SS[cr]           $       command frame start                            /
//                      AA      unit address                                   /
//                      07      command code                                   /
//                      SS      checksum (if enabled)                          /
//                      [cr]    terminator                                     /
//                                                                             /
//!AA07rSS[cr]          AA      unit address                                   /
//                      07      command code                                   /
//                      r       result a=ack/f=frame error                     /
//                      SS      checksum (if enabled)                          /
//                      [cr]    terminator                                     /
////////////////////////////////////////////////////////////////////////////////
//responses at this level                                                      /
//!AAccaSS[cr]          command accepted (no return data)                      /
//!AAccdd...ddSS[cr]    command accepted (return data)                         /
//!AAoSS[cr]            command buffer overrun error                           /
//!AAfSS[cr]            framing error (odd number of param)                    /
//!AAcSS[cr]            checksum error                                         /
////////////////////////////////////////////////////////////////////////////////
//responses at main level                                                      /
//!AAccaSS[cr]          command accepted (no return data)                      /
//!AAccdd...ddSS[cr]    command accepted (return data)                         /
//!AAccnSS[cr]          cammand parameters were bad                            /
//!AAccfSS[cr]          framing for this command was bad                       /
//!AAcc?SS[cr]          no matching command found                              /
////////////////////////////////////////////////////////////////////////////////




/*implementation code segments

//put this in the project header file, usually done by the project wizard
//#use rs232(baud=57600,parity=N,xmit=PIN_C6,rcv=PIN_C7,bits=8,stream=pcport)

////////////////////////////////////////////////////////////////////////////////
//put in main c file after the header file include
#define     scmdstream      pcport      //more than 1 stream, scmd uses spec'd
#define     scmduart        1           //1=rda, 2=rda2
#define     scmdbuffer      19          //default buffer size, limited by ram
byte const  scmd_c1='c';                //sw core, command 00 response data
byte const  scmd_c2='c';
byte const  scmd_c3='c';
byte const  scmd_c4='c';
byte const  scmd_v1='v';                //sw variant
byte const  scmd_v2='v';
byte const  scmd_r1='r';                //sw revision
byte const  scmd_r2='r';
#include    "scmd.c"                    //set path in project
#include    "convert.c"                 //main code, not interrupt handler
////////////////////////////////////////////////////////////////////////////////

//put in memory initialization block in main()
scmd_init(address,usechecksum);

//put this ion the main while(true) loop
protocol();

////////////////////////////////////////////////////////////////////////////////
//protool handler function for calling from main while(true) loop              /
//scmd  [0]     command data length in bytes                                   /
//      [1]     command code                                                   /
//      [2-18]  command data, binary                                           /
////////////////////////////////////////////////////////////////////////////////
void protocol(void){
//------------------------------------------------------------------------------
//overrun error clearing, perform every system loop
//change to xxx1 or xxx2 to match port used
//------------------------------------------------------------------------------
if (oerr1)                              //if overrun is true
    {                                   //  (Setup in scmd.c vs device number)
    cren1=false;                        //toggle rcsta, cren bit to clear,
    cren1=true;                         //  must match uart #
    }
//------------------------------------------------------------------------------
//incomming command handler
//------------------------------------------------------------------------------
if (scmd_ready)
    {
    switch (scmd[1])                    //handle commands
        {
        case 0x10:                      //test framing
            if (scmd[0]==exp_param_count)   
                {
                //code;
                //code;
                //code;
                scmd[0]='a';            //actual command handler code
                }
            else
                scmd[0]='f';            //bad framing response
            break;
        default:
            scmd[0]='?';                //un-handled command response
            break;
        }                               //multiple uarts: fputc('x',stream)
    if (scmd[0]>0){                     //if 0, command did the responding
        putc('!');                      //xmit response, use scmd variables
        bintohex(scmd_address,scmd_byte); //transmit address
        putc(scmd_byte[0]);
        putc(scmd_byte[1]);
        bintohex(scmd[1],scmd_byte);    //transmit command code
        putc(scmd_byte[0]);
        putc(scmd_byte[1]);
        putc(scmd[0]);                  //transmit response character
        putc(0x0d);                     //transmit terminator
        putc(0x0a);
        }
    scmd_ready=false;                   //unflag command ready
    }
//------------------------------------------------------------------------------
}
////////////////////////////////////////////////////////////////////////////////
*/






////////////////////////////////////////////////////////////////////////////////
//begin actual receive data interrupt handler                                  /
////////////////////////////////////////////////////////////////////////////////
//setup overrun flags and control addresses
//RCSTA1/2      Address to uart control bits
//cren1/2       continuous receive enable bit
//oerr1/2       Overrun error flag bits
////////////////////////////////////////////////////////////////////////////////
#if (__device__==876)|(__device__==877)|(__device__==873)|(__device__==874)
  //PIC16F876
  #byte rcsta1=0x18
  #bit cren1=rcsta1.4
  #bit oerr1=rcsta1.1
#endif
////////////////////////////////////////////////////////////////////////////////
#if (__device__==252)|(__device__==242)|(__device__==442)|(__device__==452)
  //PIC18F252,242,442,452
  #byte rcsta1=0xfab
  #bit cren1=rcsta1.4
  #bit oerr1=rcsta1.1
#endif
////////////////////////////////////////////////////////////////////////////////
#if (__device__==6520)|(__device__==620)|(__device__==6527)|(__device__==2520)|(__device__==4520)
  //PIC18Fxx20
  #byte rcsta1=0xfab
  #bit cren1=rcsta1.4
  #bit oerr1=rcsta1.1
  #byte rcsta2=0xf6b
  #bit cren2=rcsta2.4
  #bit oerr2=rcsta2.1
#endif
////////////////////////////////////////////////////////////////////////////////
#ifndef scmdbuffer
  #define scmdbuffer 19
#endif
unsigned    int8    scmd[scmdbuffer];       //command buffer
unsigned    int8    scmd_byte[2];           //byte building area
unsigned    int8    scmd_count;             //converted byte counter or response
unsigned    int8    scmd_checksum;          //running checksum
unsigned    int8    scmd_address;           //global unit address
unsigned    int1    scmd_ready;             //global cmmd flag:scmd set/main clr
unsigned    int1    scmd_building;          //command comming in
unsigned    int1    scmd_bytebuilding;      //ascii hex byte in progress
unsigned    int1    scmd_usechecksum;       //global checksum use flag
unsigned    int1    scmd_respond;           //local response ready
unsigned    int1    scmd_reset;             //software reset flag
////////////////////////////////////////////////////////////////////////////////
#if scmduart==1                         //1 is uart 1
    #int_RDA
    void RDA_isr(void) {
#else
    #int_RDA2                           //2 is uart 2
    void RDA2_isr(void) {
#endif
byte temp;
do {
////////////////////////////////////////////////////////////////////////////////
#ifdef scmdstream                       //if inter called, byte must be ready!
    temp=fgetc(scmdstream);
#else
    temp=getc();
#endif

if (!scmd_ready)                        //if cmmd is in queue, dump until clear
    {
    if (!scmd_building)
        {
        if (temp=='$')                  //start new command building
            {
            scmd_building=true;         //if '$'+not build, init/flag as build
            scmd_bytebuilding=false;
            scmd_checksum='$';
            scmd_count=0;
            }
        }
    else                                //is building now
        {
        switch (temp)                   //if '$' and already build, zero count
            {
            case '$':                   //restart
                scmd_bytebuilding=false;
                scmd_checksum='$';
                scmd_count=0;
                break;
            case 0x0d:                  //if [cr] and already build, terminate
                scmd_building=false;
                if (scmd_count>0)       //addr at least? else no error can be tx
                    {
                    if (scmd[0]==scmd_address)  //addr test, if not mine: dump
                        {
                        if (scmd_bytebuilding || (scmd_count==1))
                            {                   //if middle of byte, bad framing
                            scmd_respond=true;  //or only an address was sent
                            scmd_count='f';     //respond bad frameing
                            }
                        else                    //not bad framing
                            {
                            if (!scmd_usechecksum) //checksum test not enabled
                                {
                                scmd[0]=scmd_count-2; //let separate command
                                scmd_ready=true;      //processors do framing
                                }
                            else                //checksum testing enabled
                                {               //sub checksum ascii before test
                                scmd_checksum=scmd_checksum-
                                    scmd_byte[0]-scmd_byte[1];
                                if (scmd_checksum==scmd[scmd_count-1])
                                    {
                                    scmd[0]=scmd_count-3;//store parameter count
                                    scmd_ready=true;    //flag command as ready
                                    }
                                else                    //else resp bad checksum
                                    {
                                    scmd_respond=true;
                                    scmd_count='c';
                                    }
                                }
                            }
                        }
                    }
                break;
            default:        //handle char, conv ascii hex to bin as they come in
                scmd_checksum=scmd_checksum+temp; //add to csum as they come in
                if (!scmd_bytebuilding)
                    {
                    scmd_bytebuilding=true;       //flag first half done
                    scmd_byte[0]=temp;            //save first half of ascii hex
                    }
                else
                    {
                    scmd_bytebuilding=false;      //flag second half done
                    scmd_byte[1]=temp;            //save second half
                    temp=scmd_byte[0];            //convert to binary
                    swap(temp);
                    temp=temp & 0xf0;             //first half
                    if (bit_test(scmd_byte[0],6))
                        temp=temp+0x90;
                    temp=temp+(scmd_byte[1] & 0x0f); //second half
                    if (bit_test(scmd_byte[1],6))
                        temp=temp+0x09;
                    scmd[scmd_count]=temp;        //save result
                    scmd_count++;                 //incr converted byte counter
                    if (scmd_count>=scmdbuffer)   //test for buffer overrun
                        {
                        scmd_building=false;      //if overrun, dump this cmmd
                        if (scmd[0]==scmd_address)
                            {
                            scmd_respond=true;
                            scmd_count='o';
                            }
                        }
                    }
            }
        }
    }
////////////////////////////////////////////////////////////////////////////////
//low command (0x00 - 0x07 processing block                                    /
////////////////////////////////////////////////////////////////////////////////
if (scmd_ready)
    {
    switch (scmd[1])
        {
        case 0x00:                      //report firmware code
            scmd_respond=true;
            if (scmd[0]==0)             //only if data length was 0
                {
                scmd_count=scmd_r2;
                }
            else
                scmd_count='f';
            break;
        case 0x01:                      //change address
            scmd_respond=true;
            if (scmd[0]==1)             //only if data length was 1
                {
                scmd_address=scmd[2];
                scmd_count='a';
                }
            else
                scmd_count='f';
            break;
        case 0x02:                      //toggle checksum testing
            scmd_respond=true;
            if (scmd[0]==0)
                {
                scmd_usechecksum=!scmd_usechecksum;
                scmd_count='a';
                }
            else
                scmd_count='f';
            break;
        case 0x07:                      //software reset
            scmd_respond=true;
            if (scmd[0]==0)
                {
                scmd_reset=true;
                scmd_count='a';
                }
            else
                scmd_count='f';
            break;
        }
    }
////////////////////////////////////////////////////////////////////////////////
//response block, low commands (0x00 - 0x07) and errors                        /
////////////////////////////////////////////////////////////////////////////////
if (scmd_respond)                       //if flagged to respond
    {
    scmd_respond=false;                 //unflag
    #ifdef scmdstream
        fputc('!',scmdstream);          //frame start
    #else
        putc('!');                      //frame start
    #endif
    temp=scmd_address;                  //first half of unit address
    temp=(swap(temp) & 0x0f)+0x30;
    if (temp>0x39)
        temp=temp+0x07;
    scmd_checksum='!'+temp;
    #ifdef scmdstream
        fputc(temp,scmdstream);
    #else
        putc(temp);
    #endif
    temp=(scmd_address & 0x0f)+0x30;    //second half of unit address
    if (temp>0x39)
        temp=temp+0x07;
    #ifdef scmdstream
        fputc(temp,scmdstream);         //frame start
    #else
        putc(temp);                     //frame start
    #endif
    scmd_checksum=scmd_checksum+temp;
////////////////////////////////////////////////////////////////////////////////
    if (scmd_ready)                     //process extra data for low commands
        {
        scmd_ready=false;
        #ifdef scmdstream               //command always "0x" at this level
            fputc('0',scmdstream);      //frame start
            fputc(scmd[1]+0x30,scmdstream);  //second byte is ascii of cmmd code
        #else
            putc('0');                  //frame start
            putc(scmd[1]+0x30);         //second byte is ascii of command code
        #endif
        scmd_checksum=scmd_checksum+'0'+scmd[1]+'0';  //add cmmd ascii to chksum
        if ((scmd[1]==0) && (scmd[0]==0)) //if command was 0 and framing
            {                             //was good then report firmware
            #ifdef scmdstream
                fputc(scmd_c1,scmdstream);  //core
                fputc(scmd_c2,scmdstream);
                fputc(scmd_c3,scmdstream);
                fputc(scmd_c4,scmdstream);
                fputc(scmd_v1,scmdstream);  //variation
                fputc(scmd_v2,scmdstream);
                fputc(scmd_r1,scmdstream);  //revision (r2 passed in scmd_count)
            #else
                putc(scmd_c1);              //core
                putc(scmd_c2);
                putc(scmd_c3);
                putc(scmd_c4);
                putc(scmd_v1);              //variation
                putc(scmd_v2);
                putc(scmd_r1);              //revision (r2 passed in scmd_count)
            #endif
            scmd_checksum=scmd_checksum+(scmd_c1+scmd_c2+scmd_c3+
                scmd_c4+scmd_v1+scmd_v2+scmd_r1);
            }
        }
////////////////////////////////////////////////////////////////////////////////
    #ifdef scmdstream
        fputc(scmd_count,scmdstream);        //transmit parameter
    #else
        putc(scmd_count);                    //transmit parameter
    #endif
    scmd_checksum=scmd_checksum+scmd_count;
    if (scmd_usechecksum)                    //transmit checksum
        {
        temp=scmd_checksum;                  //first half of checksum
        temp=(swap(temp) & 0x0f)+0x30;
        if (temp>0x39)
            temp=temp+0x07;
        #ifdef scmdstream
            fputc(temp,scmdstream);
        #else
            putc(temp);
        #endif
        temp=(scmd_checksum & 0x0f)+0x30;    //second half of checksum
        if (temp>0x39)
            temp=temp+0x07;
        #ifdef scmdstream
            fputc(temp,scmdstream);
        #else
            putc(temp);
        #endif
        }
    putc(0x0d);                         //response terminator
    putc(0x0a);
    if (scmd_reset)                     //command requested processor reset
        {
        putc(0x00);                     //send null pads to allow last [cr] out
        putc(0x00);
        scmd_reset=false;
        reset_cpu();
        }
    }
}
//------------------------------------------------------------------------------
#ifdef scmdstream
    while (kbhit(scmdstream)==true);
#else
    while (kbhit()==true);
#endif
//------------------------------------------------------------------------------
}
////////////////////////////////////////////////////////////////////////////////
void scmd_init (unsigned int8 address,unsigned int1 checksum){
unsigned int8 i;

for (i=0;i<scmdbuffer+1;i++)
    scmd[i]=0;

scmd_count=0;
scmd_checksum=checksum;
scmd_address=address;
scmd_ready=0;
scmd_building=0;
scmd_bytebuilding=0;
scmd_usechecksum=0;
scmd_respond=0;
scmd_reset=0;
}
////////////////////////////////////////////////////////////////////////////////

